"""Package to house various prediction utilies."""

__all__ = [
    "convert_outputs",
    "convert_outputs_microsplit",
    "stitch_prediction",
    "stitch_prediction_single",
    "stitch_prediction_vae",
]

from .prediction_outputs import convert_outputs, convert_outputs_microsplit
from .stitch_prediction import (
    stitch_prediction,
    stitch_prediction_single,
    stitch_prediction_vae,
)



from collections.abc import Sequence
from enum import Enum
from pathlib import Path
from typing import Any, Generic, Literal, NamedTuple, Union

import numpy as np
from numpy.typing import NDArray
from torch.utils.data import Dataset
from tqdm.auto import tqdm

from careamics.config.data.ng_data_model import NGDataConfig
from careamics.config.support.supported_patching_strategies import (
    SupportedPatchingStrategy,
)
from careamics.config.transformations import NormalizeModel
from careamics.dataset.dataset_utils.running_stats import WelfordStatistics
from careamics.dataset.patching.patching import Stats
from careamics.dataset_ng.patch_extractor import GenericImageStack, PatchExtractor
from careamics.dataset_ng.patching_strategies import (
    FixedRandomPatchingStrategy,
    PatchingStrategy,
    PatchSpecs,
    RandomPatchingStrategy,
    TilingStrategy,
    WholeSamplePatchingStrategy,
)
from careamics.transforms import Compose


class Mode(str, Enum):
    TRAINING = "training"
    VALIDATING = "validating"
    PREDICTING = "predicting"


class ImageRegionData(NamedTuple):
    data: NDArray
    source: Union[str, Literal["array"]]
    data_shape: Sequence[int]
    dtype: str  # dtype should be str for collate
    axes: str
    region_spec: PatchSpecs


InputType = Union[Sequence[NDArray[Any]], Sequence[Path]]


class CareamicsDataset(Dataset, Generic[GenericImageStack]):
    def __init__(
        self,
        data_config: NGDataConfig,
        mode: Mode,
        input_extractor: PatchExtractor[GenericImageStack],
        target_extractor: PatchExtractor[GenericImageStack] | None = None,
    ):
        self.config = data_config
        self.mode = mode

        self.input_extractor = input_extractor
        self.target_extractor = target_extractor

        self.patching_strategy = self._initialize_patching_strategy()

        self.input_stats, self.target_stats = self._initialize_statistics()

        self.transforms = self._initialize_transforms()

    def _initialize_patching_strategy(self) -> PatchingStrategy:
        patching_strategy: PatchingStrategy
        if self.mode == Mode.TRAINING:
            if self.config.patching.name != SupportedPatchingStrategy.RANDOM:
                raise ValueError(
                    f"Only `random` patching strategy supported during training, got "
                    f"{self.config.patching.name}."
                )

            patching_strategy = RandomPatchingStrategy(
                data_shapes=self.input_extractor.shape,
                patch_size=self.config.patching.patch_size,
                seed=self.config.seed,
            )
        elif self.mode == Mode.VALIDATING:
            if self.config.patching.name != SupportedPatchingStrategy.RANDOM:
                raise ValueError(
                    f"Only `random` patching strategy supported during training, got "
                    f"{self.config.patching.name}."
                )

            patching_strategy = FixedRandomPatchingStrategy(
                data_shapes=self.input_extractor.shape,
                patch_size=self.config.patching.patch_size,
                seed=self.config.seed,
            )
        elif self.mode == Mode.PREDICTING:
            if (
                self.config.patching.name != SupportedPatchingStrategy.TILED
                and self.config.patching.name != SupportedPatchingStrategy.WHOLE
            ):
                raise ValueError(
                    f"Only `tiled` and `whole` patching strategy supported during "
                    f"training, got {self.config.patching.name}."
                )
            elif self.config.patching.name == SupportedPatchingStrategy.TILED:
                patching_strategy = TilingStrategy(
                    data_shapes=self.input_extractor.shape,
                    tile_size=self.config.patching.patch_size,
                    overlaps=self.config.patching.overlaps,
                )
            else:
                patching_strategy = WholeSamplePatchingStrategy(
                    data_shapes=self.input_extractor.shape
                )
        else:
            raise ValueError(f"Unrecognised dataset mode {self.mode}.")

        return patching_strategy

    def _initialize_transforms(self) -> Compose | None:
        normalize = NormalizeModel(
            image_means=self.input_stats.means,
            image_stds=self.input_stats.stds,
            target_means=self.target_stats.means,
            target_stds=self.target_stats.stds,
        )
        if self.mode == Mode.TRAINING:
            # TODO: initialize normalization separately depending on configuration
            return Compose(transform_list=[normalize] + list(self.config.transforms))

        # TODO: add TTA
        return Compose(transform_list=[normalize])

    def _calculate_stats(
        self, data_extractor: PatchExtractor[GenericImageStack]
    ) -> Stats:
        image_stats = WelfordStatistics()
        n_patches = self.patching_strategy.n_patches

        for idx in tqdm(range(n_patches), desc="Computing statistics"):
            patch_spec = self.patching_strategy.get_patch_spec(idx)
            patch = data_extractor.extract_patch(
                data_idx=patch_spec["data_idx"],
                sample_idx=patch_spec["sample_idx"],
                coords=patch_spec["coords"],
                patch_size=patch_spec["patch_size"],
            )
            # TODO: statistics accept SCYX format, while patch is CYX
            image_stats.update(patch[None, ...], sample_idx=idx)

        image_means, image_stds = image_stats.finalize()
        return Stats(image_means, image_stds)

    # TODO: add running stats
    def _initialize_statistics(self) -> tuple[Stats, Stats]:
        if self.config.image_means is not None and self.config.image_stds is not None:
            input_stats = Stats(self.config.image_means, self.config.image_stds)
        else:
            input_stats = self._calculate_stats(self.input_extractor)

        target_stats = Stats((), ())

        if self.config.target_means is not None and self.config.target_stds is not None:
            target_stats = Stats(self.config.target_means, self.config.target_stds)
        elif self.target_extractor is not None:
            target_stats = self._calculate_stats(self.target_extractor)

        return input_stats, target_stats

    def __len__(self):
        return self.patching_strategy.n_patches

    def _create_image_region(
        self, patch: np.ndarray, patch_spec: PatchSpecs, extractor: PatchExtractor
    ) -> ImageRegionData:
        data_idx = patch_spec["data_idx"]
        source = extractor.image_stacks[data_idx].source
        return ImageRegionData(
            data=patch,
            source=str(source),
            dtype=str(extractor.image_stacks[data_idx].data_dtype),
            data_shape=extractor.image_stacks[data_idx].data_shape,
            # TODO: should it be axes of the original image instead?
            axes=self.config.axes,
            region_spec=patch_spec,
        )

    def __getitem__(
        self, index: int
    ) -> Union[tuple[ImageRegionData], tuple[ImageRegionData, ImageRegionData]]:
        patch_spec = self.patching_strategy.get_patch_spec(index)
        input_patch = self.input_extractor.extract_patch(
            data_idx=patch_spec["data_idx"],
            sample_idx=patch_spec["sample_idx"],
            coords=patch_spec["coords"],
            patch_size=patch_spec["patch_size"],
        )

        target_patch = (
            self.target_extractor.extract_patch(
                data_idx=patch_spec["data_idx"],
                sample_idx=patch_spec["sample_idx"],
                coords=patch_spec["coords"],
                patch_size=patch_spec["patch_size"],
            )
            if self.target_extractor is not None
            else None
        )

        if self.transforms is not None:
            if self.target_extractor is not None:
                input_patch, target_patch = self.transforms(input_patch, target_patch)
            else:
                # TODO: compose doesn't return None for target patch anymore
                #   so have to do this annoying if else
                (input_patch,) = self.transforms(input_patch, target_patch)
                target_patch = None

        input_data = self._create_image_region(
            patch=input_patch, patch_spec=patch_spec, extractor=self.input_extractor
        )

        if target_patch is not None and self.target_extractor is not None:
            target_data = self._create_image_region(
                patch=target_patch,
                patch_spec=patch_spec,
                extractor=self.target_extractor,
            )
            return input_data, target_data
        else:
            return (input_data,)



from contextlib import nullcontext
from pathlib import Path
from typing import Literal

import numpy as np
import pytest
from numpy.typing import NDArray

from careamics.dataset_ng.patch_extractor.image_stack import CziImageStack

# skip if fail imports
pylib = pytest.importorskip("pylibCZIrw")

from pylibCZIrw import czi as pyczi  # noqa: E402

T_EXPR = "does not contain a T axis"
Z_EXPR = "does not contain a Z axis"


def create_test_czi(file_path: Path, data: NDArray | list[NDArray]):
    if not isinstance(data, list):
        data = [data]

    with pyczi.create_czi(str(file_path)) as czi:
        xoffs = 0
        for scene_idx, scene_data in enumerate(data):
            for t in range(scene_data.shape[0]):
                for c in range(scene_data.shape[1]):
                    for z in range(scene_data.shape[2]):
                        czi.write(
                            scene_data[t, c, z],
                            plane={"C": c, "T": t, "Z": z},
                            location=(xoffs, 0),
                            scene=scene_idx,
                        )
            xoffs += scene_data.shape[-1] + 20


@pytest.mark.czi
@pytest.mark.parametrize(
    "orig_shape, depth_axis, expected_axes, expected_shape, sample_idx, expect_raise",
    [
        # 2-D Images
        ((1, 1, 1, 32, 48), "none", "SCYX", [1, 1, 32, 48], 0, nullcontext()),
        ((1, 1, 1, 32, 48), "T", "", [], 0, pytest.raises(RuntimeError, match=T_EXPR)),
        ((1, 1, 1, 32, 48), "Z", "", [], 0, pytest.raises(RuntimeError, match=Z_EXPR)),
        # 3-D Volumes
        ((1, 1, 16, 32, 48), "none", "SCYX", [16, 1, 32, 48], 9, nullcontext()),
        ((1, 1, 16, 32, 48), "T", "", [], 9, pytest.raises(RuntimeError, match=T_EXPR)),
        ((1, 1, 16, 32, 48), "Z", "SCZYX", [1, 1, 16, 32, 48], 0, nullcontext()),
        # 2-D Time-Series
        ((8, 1, 1, 32, 48), "none", "SCYX", [8, 1, 32, 48], 3, nullcontext()),
        ((8, 1, 1, 32, 48), "Z", "", [], 3, pytest.raises(RuntimeError, match=Z_EXPR)),
        ((8, 1, 1, 32, 48), "T", "SCTYX", [1, 1, 8, 32, 48], 0, nullcontext()),
        # 3-D Time-Series
        ((8, 1, 16, 32, 48), "none", "SCYX", [8 * 16, 1, 32, 48], 35, nullcontext()),
        ((8, 1, 16, 32, 48), "Z", "SCZYX", [8, 1, 16, 32, 48], 7, nullcontext()),
        ((8, 1, 16, 32, 48), "T", "SCTYX", [16, 1, 8, 32, 48], 12, nullcontext()),
        # Multiple channels
        ((8, 3, 16, 32, 48), "Z", "SCZYX", [8, 3, 16, 32, 48], 2, nullcontext()),
    ],
)
def test_extract_patch(
    tmp_path: Path,
    orig_shape: tuple[int, ...],
    depth_axis: Literal["none", "Z", "T"],
    expected_axes: str,
    expected_shape: list[int],
    sample_idx: int,
    expect_raise,
):
    # reference data to compare against
    data = np.random.randn(*orig_shape).astype(np.float32)

    # save data as a czi file to ininitialise image stack with
    file_path = tmp_path / "test_czi.czi"
    create_test_czi(file_path=file_path, data=data)

    # initialise CziImageStack
    with expect_raise:
        image_stack = CziImageStack(data_path=file_path, depth_axis=depth_axis)

    # stop here if expecting an exception
    if expected_axes == "":
        return

    # check axes and shape
    assert image_stack.axes == expected_axes
    assert image_stack.data_shape == expected_shape

    # test extracted patch matches patch from reference data
    if len(expected_axes) < 5:
        coords = (11, 4)
        patch_size = (16, 9)

        extracted_patch = image_stack.extract_patch(
            sample_idx=sample_idx, coords=coords, patch_size=patch_size
        )

        data_ref = np.moveaxis(data, 2, 1)  # (T, Z, C, Y, X)
        data_ref = data_ref.reshape(-1, *data_ref.shape[-3:])
        patch_ref = data_ref[
            sample_idx,
            :,
            coords[0] : coords[0] + patch_size[0],
            coords[1] : coords[1] + patch_size[1],
        ]
    else:
        coords = (2, 11, 4)
        patch_size = (4, 16, 9)

        extracted_patch = image_stack.extract_patch(
            sample_idx=sample_idx, coords=coords, patch_size=patch_size
        )

        if "T" in expected_axes and expected_axes.index("T") == 2:
            data_ref = data.swapaxes(0, 2)
        else:
            data_ref = data
        patch_ref = data_ref[
            sample_idx,
            ...,
            coords[0] : coords[0] + patch_size[0],
            coords[1] : coords[1] + patch_size[1],
            coords[2] : coords[2] + patch_size[2],
        ]
    np.testing.assert_array_equal(extracted_patch, patch_ref)


@pytest.mark.czi
def test_multiple_scenes(
    tmp_path: Path,
):
    original_shapes = [
        [4, 1, 8, 16, 32],
        [4, 1, 8, 24, 18],
        [4, 1, 8, 45, 32],
    ]

    # reference data to compare against
    data_ref = [np.random.randn(*shape).astype(np.float32) for shape in original_shapes]

    # save data as a czi file to ininitialise image stack with
    file_path = tmp_path / "test_czi.czi"
    create_test_czi(file_path=file_path, data=data_ref)

    # Test reading scene metadata
    scene_rectangles = CziImageStack.get_bounding_rectangles(file_path)
    assert len(scene_rectangles) == len(original_shapes)

    scene_rectangle_sizes = [(rect.h, rect.w) for rect in scene_rectangles.values()]
    expected_rectangle_sizes = [(shape[-2], shape[-1]) for shape in original_shapes]
    assert scene_rectangle_sizes == expected_rectangle_sizes

    # Test reading from individual scenes
    for scene_idx, expected_shape in enumerate(original_shapes):
        image_stack = CziImageStack(
            data_path=file_path, scene=scene_idx, depth_axis="Z"
        )
        assert image_stack.data_shape == expected_shape

        t = expected_shape[0] - scene_idx - 1
        coords = (2, 9, 4)
        patch_size = (4, 7, 13)

        extracted_patch = image_stack.extract_patch(
            sample_idx=t, coords=coords, patch_size=patch_size
        )
        patch_ref = data_ref[scene_idx][
            t,
            :,
            coords[0] : coords[0] + patch_size[0],
            coords[1] : coords[1] + patch_size[1],
            coords[2] : coords[2] + patch_size[2],
        ]
        np.testing.assert_array_equal(extracted_patch, patch_ref)



"""CARE Lightning DataModule."""

from collections.abc import Callable
from typing import Any, Union

from careamics.config.algorithms.care_algorithm_model import CAREAlgorithm
from careamics.config.algorithms.n2n_algorithm_model import N2NAlgorithm
from careamics.config.support import SupportedLoss
from careamics.dataset_ng.dataset import ImageRegionData
from careamics.losses import mae_loss, mse_loss
from careamics.utils.logging import get_logger

from .unet_module import UnetModule

logger = get_logger(__name__)


class CAREModule(UnetModule):
    """CAREamics PyTorch Lightning module for CARE algorithm.

    Parameters
    ----------
    algorithm_config : CAREAlgorithm or dict
        Configuration for the CARE algorithm, either as a CAREAlgorithm instance or a
        dictionary.
    """

    def __init__(self, algorithm_config: Union[CAREAlgorithm, dict]) -> None:
        """Instantiate CARE DataModule.

        Parameters
        ----------
        algorithm_config : CAREAlgorithm or dict
            Configuration for the CARE algorithm, either as a CAREAlgorithm instance or
            a dictionary.
        """
        super().__init__(algorithm_config)
        assert isinstance(
            algorithm_config, CAREAlgorithm | N2NAlgorithm
        ), "algorithm_config must be a CAREAlgorithm or a N2NAlgorithm"
        loss = algorithm_config.loss
        if loss == SupportedLoss.MAE:
            self.loss_func: Callable = mae_loss
        elif loss == SupportedLoss.MSE:
            self.loss_func = mse_loss
        else:
            raise ValueError(f"Unsupported loss for Care: {loss}")

    def training_step(
        self,
        batch: tuple[ImageRegionData, ImageRegionData],
        batch_idx: Any,
    ) -> Any:
        """Training step for CARE module.

        Parameters
        ----------
        batch : (ImageRegionData, ImageRegionData)
            A tuple containing the input data and the target data.
        batch_idx : Any
            The index of the current batch in the training loop.

        Returns
        -------
        Any
            The loss value computed for the current batch.
        """
        # TODO: add validation to determine if target is initialized
        x, target = batch[0], batch[1]

        prediction = self.model(x.data)
        loss = self.loss_func(prediction, target.data)

        self._log_training_stats(loss, batch_size=x.data.shape[0])

        return loss

    def validation_step(
        self,
        batch: tuple[ImageRegionData, ImageRegionData],
        batch_idx: Any,
    ) -> None:
        """Validation step for CARE module.

        Parameters
        ----------
        batch : (ImageRegionData, ImageRegionData)
            A tuple containing the input data and the target data.
        batch_idx : Any
            The index of the current batch in the training loop.
        """
        x, target = batch[0], batch[1]

        prediction = self.model(x.data)
        val_loss = self.loss_func(prediction, target.data)
        self.metrics(prediction, target.data)
        self._log_validation_stats(val_loss, batch_size=x.data.shape[0])



from collections.abc import Sequence
from typing import Generic

from numpy.typing import NDArray

from .image_stack import GenericImageStack


class PatchExtractor(Generic[GenericImageStack]):
    """
    A class for extracting patches from multiple image stacks.
    """

    def __init__(self, image_stacks: Sequence[GenericImageStack]):
        self.image_stacks: list[GenericImageStack] = list(image_stacks)

    def extract_patch(
        self,
        data_idx: int,
        sample_idx: int,
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        return self.image_stacks[data_idx].extract_patch(
            sample_idx=sample_idx, coords=coords, patch_size=patch_size
        )

    @property
    def shape(self):
        return [stack.data_shape for stack in self.image_stacks]



"""MicroSplit algorithm configuration."""

from typing import Literal

from bioimageio.spec.generic.v0_3 import CiteEntry
from pydantic import ConfigDict

from careamics.config.algorithms.vae_algorithm_model import VAEBasedAlgorithm
from careamics.config.architectures import LVAEModel
from careamics.config.loss_model import LVAELossConfig

MICROSPLIT = "MicroSplit"

MICROSPLIT_DESCRIPTION = """MicroSplit is a self-supervised deep learning method for
microscopy image splitting that combines the strengths of both denoising and
representation learning approaches."""

MICROSPLIT_REF = CiteEntry(
    text='Prakash, M., Delbracio, M., Milanfar, P., Jug, F. 2022. "Interpretable '
    'Unsupervised Diversity Denoising and Artefact Removal." The International '
    "Conference on Learning Representations (ICLR).",
    doi="10.1561/2200000056",
)


class MicroSplitAlgorithm(VAEBasedAlgorithm):
    """MicroSplit algorithm configuration."""

    model_config = ConfigDict(validate_assignment=True)

    algorithm: Literal["microsplit"] = "microsplit"

    loss: LVAELossConfig

    model: LVAEModel  # TODO add validators

    is_supervised: bool = True

    def get_algorithm_friendly_name(self) -> str:
        """
        Get the algorithm friendly name.

        Returns
        -------
        str
            Friendly name of the algorithm.
        """
        return MICROSPLIT

    def get_algorithm_keywords(self) -> list[str]:
        """
        Get algorithm keywords.

        Returns
        -------
        list[str]
            List of keywords.
        """
        return [
            "restoration",
            "VAE",
            "self-supervised",
            "3D" if self.model.is_3D() else "2D",
            "CAREamics",
            "pytorch",
        ]

    def get_algorithm_references(self) -> str:
        """
        Get the algorithm references.

        This is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Algorithm references.
        """
        return MICROSPLIT_REF.text + " doi: " + MICROSPLIT_REF.doi

    def get_algorithm_citations(self) -> list[CiteEntry]:
        """
        Return a list of citation entries of the current algorithm.

        This is used to generate the model description for the BioImage Model Zoo.

        Returns
        -------
        List[CiteEntry]
            List of citation entries.
        """
        return [MICROSPLIT_REF]

    def get_algorithm_description(self) -> str:
        """
        Get the algorithm description.

        Returns
        -------
        str
            Algorithm description.
        """
        return MICROSPLIT_DESCRIPTION



"""Generic UNet Lightning DataModule."""

from typing import Any, Union

import pytorch_lightning as L
import torch
from torch import nn
from torchmetrics import MetricCollection
from torchmetrics.image import PeakSignalNoiseRatio

from careamics.config import algorithm_factory
from careamics.config.algorithms import CAREAlgorithm, N2NAlgorithm, N2VAlgorithm
from careamics.dataset_ng.dataset import ImageRegionData
from careamics.models.unet import UNet
from careamics.transforms import Denormalize
from careamics.utils.logging import get_logger
from careamics.utils.torch_utils import get_optimizer, get_scheduler

logger = get_logger(__name__)


class UnetModule(L.LightningModule):
    """CAREamics PyTorch Lightning module for UNet based algorithms.

    Parameters
    ----------
    algorithm_config : CAREAlgorithm, N2VAlgorithm, N2NAlgorithm, or dict
        Configuration for the algorithm, either as an instance of a specific algorithm
        class or a dictionary that can be converted to an algorithm instance.
    """

    def __init__(
        self, algorithm_config: Union[CAREAlgorithm, N2VAlgorithm, N2NAlgorithm, dict]
    ) -> None:
        """Instantiate UNet DataModule.

        Parameters
        ----------
        algorithm_config : CAREAlgorithm, N2VAlgorithm, N2NAlgorithm, or dict
            Configuration for the algorithm, either as an instance of a specific
            algorithm class or a dictionary that can be converted to an algorithm
            instance.
        """
        super().__init__()

        if isinstance(algorithm_config, dict):
            algorithm_config = algorithm_factory(algorithm_config)

        self.config = algorithm_config
        self.model: nn.Module = UNet(**algorithm_config.model.model_dump())

        self._best_checkpoint_loaded = False

        # TODO: how to support metric evaluation better
        self.metrics = MetricCollection(PeakSignalNoiseRatio())

    def forward(self, x: Any) -> Any:
        """Default forward method.

        Parameters
        ----------
        x : Any
            Input data.

        Returns
        -------
        Any
            Output from the model.
        """
        return self.model(x)

    def _log_training_stats(self, loss: Any, batch_size: Any) -> None:
        """Log training statistics.

        Parameters
        ----------
        loss : Any
            The loss value for the current training step.
        batch_size : Any
            The size of the batch used in the current training step.
        """
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            batch_size=batch_size,
        )

        optimizer = self.optimizers()
        if isinstance(optimizer, list):
            current_lr = optimizer[0].param_groups[0]["lr"]
        else:
            current_lr = optimizer.param_groups[0]["lr"]
        self.log(
            "learning_rate",
            current_lr,
            on_step=False,
            on_epoch=True,
            logger=True,
            batch_size=batch_size,
        )

    def _log_validation_stats(self, loss: Any, batch_size: Any) -> None:
        """Log validation statistics.

        Parameters
        ----------
        loss : Any
            The loss value for the current validation step.
        batch_size : Any
            The size of the batch used in the current validation step.
        """
        self.log(
            "val_loss",
            loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            batch_size=batch_size,
        )
        self.log_dict(self.metrics, on_step=False, on_epoch=True, batch_size=batch_size)

    def _load_best_checkpoint(self) -> None:
        """Load the best checkpoint from the trainer's checkpoint callback."""
        if (
            not hasattr(self.trainer, "checkpoint_callback")
            or self.trainer.checkpoint_callback is None
        ):
            logger.warning("No checkpoint callback found, cannot load best checkpoint.")
            return

        best_model_path = self.trainer.checkpoint_callback.best_model_path
        if best_model_path and best_model_path != "":
            logger.info(f"Loading best checkpoint from: {best_model_path}")
            model_state = torch.load(best_model_path, weights_only=True)["state_dict"]
            self.load_state_dict(model_state)
        else:
            logger.warning("No best checkpoint found.")

    def predict_step(
        self,
        batch: Union[tuple[ImageRegionData], tuple[ImageRegionData, ImageRegionData]],
        batch_idx: Any,
        load_best_checkpoint=False,
    ) -> Any:
        """Default predict step.

        Parameters
        ----------
        batch : ImageRegionData or (ImageRegionData, ImageRegionData)
            A tuple containing the input data and optionally the target data.
        batch_idx : Any
            The index of the current batch in the prediction loop.
        load_best_checkpoint : bool, default=False
            Whether to load the best checkpoint before making predictions.

        Returns
        -------
        Any
            The output batch containing the predictions.
        """
        if self._best_checkpoint_loaded is False and load_best_checkpoint:
            self._load_best_checkpoint()
            self._best_checkpoint_loaded = True

        x = batch[0]
        # TODO: add TTA
        prediction = self.model(x.data).cpu().numpy()

        means = self._trainer.datamodule.stats.means
        stds = self._trainer.datamodule.stats.stds
        denormalize = Denormalize(
            image_means=means,
            image_stds=stds,
        )
        denormalized_output = denormalize(prediction)

        output_batch = ImageRegionData(
            data=denormalized_output,
            source=x.source,
            data_shape=x.data_shape,
            dtype=x.dtype,
            axes=x.axes,
            region_spec=x.region_spec,
        )
        return output_batch

    def configure_optimizers(self) -> Any:
        """Configure optimizers.

        Returns
        -------
        Any
            A dictionary containing the optimizer and learning rate scheduler.
        """
        optimizer_func = get_optimizer(self.config.optimizer.name)
        optimizer = optimizer_func(
            self.model.parameters(), **self.config.optimizer.parameters
        )

        scheduler_func = get_scheduler(self.config.lr_scheduler.name)
        scheduler = scheduler_func(optimizer, **self.config.lr_scheduler.parameters)

        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "val_loss",  # otherwise triggers MisconfigurationException
        }



"""Data statistics callback."""

import pytorch_lightning as L
from pytorch_lightning.callbacks import Callback


class DataStatsCallback(Callback):
    """Callback to update model's data statistics from datamodule.

    This callback ensures that the model has access to the data statistics (mean and std)
    calculated by the datamodule before training starts.
    """

    def setup(self, trainer: L.Trainer, module: L.LightningModule, stage: str) -> None:
        """Called when trainer is setting up."""
        if stage == "fit":
            # Get data statistics from datamodule
            (data_mean, data_std), _ = trainer.datamodule.get_data_stats()

            # Set data statistics in the model's likelihood module
            module.noise_model_likelihood.set_data_stats(
                data_mean=data_mean["target"], data_std=data_std["target"]
            )



"""Random patching Pydantic model."""

from typing import Literal

from ._patched_model import _PatchedModel


class RandomPatchingModel(_PatchedModel):
    """Random patching Pydantic model.

    Attributes
    ----------
    name : "random"
        The name of the patching strategy.
    patch_size : sequence of int
        The size of the patch in each spatial dimension, each patch size must be a power
        of 2 and larger than 8.
    """

    name: Literal["random"] = "random"
    """The name of the patching strategy."""



__all__ = ["GenericImageStack", "ImageStackLoader", "PatchExtractor"]

from .image_stack import GenericImageStack
from .image_stack_loader import ImageStackLoader
from .patch_extractor import PatchExtractor



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal

from numpy.typing import NDArray
from typing_extensions import ParamSpec

from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.file_io.read import ReadFunc

from .image_stack import (
    CziImageStack,
    GenericImageStack,
    InMemoryImageStack,
    ZarrImageStack,
)
from .image_stack_loader import (
    ImageStackLoader,
)

P = ParamSpec("P")


# Array case
def create_array_extractor(
    source: Sequence[NDArray[Any]], axes: str
) -> PatchExtractor[InMemoryImageStack]:
    """
    Create a patch extractor from a sequence of numpy arrays.

    Parameters
    ----------
    source: sequence of numpy.ndarray
        The source arrays of the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    PatchExtractor
    """
    image_stacks = [
        InMemoryImageStack.from_array(data=array, axes=axes) for array in source
    ]
    return PatchExtractor(image_stacks)


# TIFF case
def create_tiff_extractor(
    source: Sequence[Path], axes: str
) -> PatchExtractor[InMemoryImageStack]:
    """
    Create a patch extractor from a sequence of TIFF files.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    PatchExtractor
    """
    image_stacks = [
        InMemoryImageStack.from_tiff(path=path, axes=axes) for path in source
    ]
    return PatchExtractor(image_stacks)


# ZARR case
def create_ome_zarr_extractor(
    source: Sequence[Path],
    axes: str,
) -> PatchExtractor[ZarrImageStack]:
    """
    Create a patch extractor from a sequence of OME ZARR files.

    If you have ZARR files that do not follow the OME standard, see documentation on
    how to create a custom `image_stack_loader`. (TODO: Add link).

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    PatchExtractor
    """
    # NOTE: axes is unused here, in from_ome_zarr the axes are automatically retrieved
    image_stacks = [ZarrImageStack.from_ome_zarr(path) for path in source]
    return PatchExtractor(image_stacks)


# CZI case
def create_czi_extractor(
    source: Sequence[Path],
    axes: str,
) -> PatchExtractor[CziImageStack]:
    """
    Create a patch extractor from a sequence of CZI files.

    If the CZI files contain multiple scenes, one patch extractor will be created for
    each scene.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        Specifies which axes of the data to use and how.
        If this string ends with `"ZYX"` or `"TYX"`, the data will consist of 3-D
        patches, using `Z` or `T` as third dimension, respectively.
        If the string does not end with "ZYX", the data will consist of 2-D patches.

    Returns
    -------
    PatchExtractor
    """
    depth_axis: Literal["none", "Z", "T"] = "none"
    if axes.endswith("TYX"):
        depth_axis = "T"
    elif axes.endswith("ZYX"):
        depth_axis = "Z"

    image_stacks: list[CziImageStack] = []
    for path in source:
        scene_rectangles = CziImageStack.get_bounding_rectangles(path)
        image_stacks.extend(
            CziImageStack(path, scene=scene, depth_axis=depth_axis)
            for scene in scene_rectangles.keys()
        )
    return PatchExtractor(image_stacks)


# Custom file type case (loaded into memory)
def create_custom_file_extractor(
    source: Sequence[Path],
    axes: str,
    *,
    read_func: ReadFunc,
    read_kwargs: dict[str, Any],
) -> PatchExtractor[InMemoryImageStack]:
    """
    Create a patch extractor from a sequence of files of a custom type.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".
    read_func : ReadFunc
        A function to read the custom file type, see the `ReadFunc` protocol.
    read_kwargs : dict of {str: Any}
        Kwargs that will be passed to the custom `read_func`.

    Returns
    -------
    PatchExtractor
    """
    # TODO: lazy loading custom files
    image_stacks = [
        InMemoryImageStack.from_custom_file_type(
            path=path,
            axes=axes,
            read_func=read_func,
            **read_kwargs,
        )
        for path in source
    ]

    return PatchExtractor(image_stacks)


# Custom ImageStackLoader case
def create_custom_image_stack_extractor(
    source: Any,
    axes: str,
    image_stack_loader: ImageStackLoader[P, GenericImageStack],
    *args: P.args,
    **kwargs: P.kwargs,
) -> PatchExtractor[GenericImageStack]:
    """
    Create a patch extractor using a custom `ImageStackLoader`.

    The custom image stack loader must follow the `ImageStackLoader` protocol, i.e.
    it must have the following function signature:
    ```
    def image_loader_example(
        source: Any, data_config: DataConfig, *args, **kwargs
    ) -> Sequence[ImageStack]:
    ```

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".
    image_stack_loader: ImageStackLoader
        A custom image stack loader callable.
    *args: Any
        Positional arguments that will be passed to the custom image stack loader.
    **kwargs: Any
        Keyword arguments that will be passed to the custom image stack loader.

    Returns
    -------
    PatchExtractor
    """
    image_stacks = image_stack_loader(source, axes, *args, **kwargs)
    return PatchExtractor(image_stacks)



"""Next-Generation CAREamics DataModule."""

from collections.abc import Callable
from pathlib import Path
from typing import Any, Union, overload

import numpy as np
import pytorch_lightning as L
from numpy.typing import NDArray
from torch.utils.data import DataLoader
from torch.utils.data._utils.collate import default_collate

from careamics.config.data.ng_data_model import NGDataConfig
from careamics.config.support import SupportedData
from careamics.dataset.dataset_utils import list_files, validate_source_target_files
from careamics.dataset_ng.dataset import Mode
from careamics.dataset_ng.factory import create_dataset
from careamics.dataset_ng.patch_extractor import ImageStackLoader
from careamics.utils import get_logger

logger = get_logger(__name__)

ItemType = Union[Path, str, NDArray[Any]]
"""Type of input items passed to the dataset."""

InputType = Union[ItemType, list[ItemType], None]
"""Type of input data passed to the dataset."""


class CareamicsDataModule(L.LightningDataModule):
    """Data module for Careamics dataset.

    Parameters
    ----------
    data_config : DataConfig
        Pydantic model for CAREamics data configuration.
    train_data : Optional[InputType]
        Training data, can be a path to a folder, a list of paths, or a numpy array.
    train_data_target : Optional[InputType]
        Training data target, can be a path to a folder,
        a list of paths, or a numpy array.
    val_data : Optional[InputType]
        Validation data, can be a path to a folder,
        a list of paths, or a numpy array.
    val_data_target : Optional[InputType]
        Validation data target, can be a path to a folder,
        a list of paths, or a numpy array.
    pred_data : Optional[InputType]
        Prediction data, can be a path to a folder, a list of paths,
        or a numpy array.
    pred_data_target : Optional[InputType]
        Prediction data target, can be a path to a folder,
        a list of paths, or a numpy array.
    read_source_func : Optional[Callable], default=None
        Function to read the source data. Only used for `custom`
        data type (see DataModel).
    read_kwargs : Optional[dict[str, Any]]
        The kwargs for the read source function.
    image_stack_loader : Optional[ImageStackLoader]
        The image stack loader.
    image_stack_loader_kwargs : Optional[dict[str, Any]]
        The image stack loader kwargs.
    extension_filter : str, default=""
        Filter for file extensions. Only used for `custom` data types
        (see DataModel).
    val_percentage : Optional[float]
        Percentage of the training data to use for validation. Only
        used if `val_data` is None.
    val_minimum_split : int, default=5
        Minimum number of patches or files to split from the training data for
        validation. Only used if `val_data` is None.
    use_in_memory : bool
        Load data in memory dataset if possible, by default True.


    Attributes
    ----------
    config : DataConfig
        Pydantic model for CAREamics data configuration.
    data_type : str
        Type of data, one of SupportedData.
    batch_size : int
        Batch size for the dataloaders.
    use_in_memory : bool
        Whether to load data in memory if possible.
    extension_filter : str
        Filter for file extensions, by default "".
    read_source_func : Optional[Callable], default=None
        Function to read the source data.
    read_kwargs : Optional[dict[str, Any]], default=None
        The kwargs for the read source function.
    val_percentage : Optional[float]
        Percentage of the training data to use for validation.
    val_minimum_split : int, default=5
        Minimum number of patches or files to split from the training data for
        validation.
    train_data : Optional[Any]
        Training data, can be a path to a folder, a list of paths, or a numpy array.
    train_data_target : Optional[Any]
        Training data target, can be a path to a folder, a list of paths, or a numpy
        array.
    val_data : Optional[Any]
        Validation data, can be a path to a folder, a list of paths, or a numpy array.
    val_data_target : Optional[Any]
        Validation data target, can be a path to a folder, a list of paths, or a numpy
        array.
    pred_data : Optional[Any]
        Prediction data, can be a path to a folder, a list of paths, or a numpy array.
    pred_data_target : Optional[Any]
        Prediction data target, can be a path to a folder, a list of paths, or a numpy
        array.

    Raises
    ------
    ValueError
        If at least one of train_data, val_data or pred_data is not provided.
    ValueError
        If input and target data types are not consistent.
    """

    # standard use
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: InputType | None = None,
        train_data_target: InputType | None = None,
        val_data: InputType | None = None,
        val_data_target: InputType | None = None,
        pred_data: InputType | None = None,
        pred_data_target: InputType | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
        use_in_memory: bool = True,
    ) -> None: ...

    # custom read function
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: InputType | None = None,
        train_data_target: InputType | None = None,
        val_data: InputType | None = None,
        val_data_target: InputType | None = None,
        pred_data: InputType | None = None,
        pred_data_target: InputType | None = None,
        read_source_func: Callable,
        read_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
        use_in_memory: bool = True,
    ) -> None: ...

    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: Any | None = None,
        train_data_target: Any | None = None,
        val_data: Any | None = None,
        val_data_target: Any | None = None,
        pred_data: Any | None = None,
        pred_data_target: Any | None = None,
        image_stack_loader: ImageStackLoader,
        image_stack_loader_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
        use_in_memory: bool = True,
    ) -> None: ...

    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: Any | None = None,
        train_data_target: Any | None = None,
        val_data: Any | None = None,
        val_data_target: Any | None = None,
        pred_data: Any | None = None,
        pred_data_target: Any | None = None,
        read_source_func: Callable | None = None,
        read_kwargs: dict[str, Any] | None = None,
        image_stack_loader: ImageStackLoader | None = None,
        image_stack_loader_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
        use_in_memory: bool = True,
    ) -> None:
        """
        Data module for Careamics dataset initialization.

        Create a lightning datamodule that handles creating datasets for training,
        validation, and prediction.

        Parameters
        ----------
        data_config : NGDataConfig
            Pydantic model for CAREamics data configuration.
        train_data : Optional[InputType]
            Training data, can be a path to a folder, a list of paths, or a numpy array.
        train_data_target : Optional[InputType]
            Training data target, can be a path to a folder,
            a list of paths, or a numpy array.
        val_data : Optional[InputType]
            Validation data, can be a path to a folder,
            a list of paths, or a numpy array.
        val_data_target : Optional[InputType]
            Validation data target, can be a path to a folder,
            a list of paths, or a numpy array.
        pred_data : Optional[InputType]
            Prediction data, can be a path to a folder, a list of paths,
            or a numpy array.
        pred_data_target : Optional[InputType]
            Prediction data target, can be a path to a folder,
            a list of paths, or a numpy array.
        read_source_func : Optional[Callable]
            Function to read the source data, by default None. Only used for `custom`
            data type (see DataModel).
        read_kwargs : Optional[dict[str, Any]]
            The kwargs for the read source function.
        image_stack_loader : Optional[ImageStackLoader]
            The image stack loader.
        image_stack_loader_kwargs : Optional[dict[str, Any]]
            The image stack loader kwargs.
        extension_filter : str
            Filter for file extensions, by default "". Only used for `custom` data types
            (see DataModel).
        val_percentage : Optional[float]
            Percentage of the training data to use for validation. Only
            used if `val_data` is None.
        val_minimum_split : int
            Minimum number of patches or files to split from the training data for
            validation, by default 5. Only used if `val_data` is None.
        use_in_memory : bool
            Load data in memory dataset if possible, by default True.
        """
        super().__init__()

        if train_data is None and val_data is None and pred_data is None:
            raise ValueError(
                "At least one of train_data, val_data or pred_data must be provided."
            )

        self.config: NGDataConfig = data_config
        self.data_type: str = data_config.data_type
        self.batch_size: int = data_config.batch_size
        self.use_in_memory: bool = use_in_memory
        self.extension_filter: str = extension_filter
        self.read_source_func = read_source_func
        self.read_kwargs = read_kwargs
        self.image_stack_loader = image_stack_loader
        self.image_stack_loader_kwargs = image_stack_loader_kwargs

        # TODO: implement the validation split logic
        self.val_percentage = val_percentage
        self.val_minimum_split = val_minimum_split
        if self.val_percentage is not None:
            raise NotImplementedError("Validation split not implemented")

        self.train_data, self.train_data_target = self._initialize_data_pair(
            train_data, train_data_target
        )
        self.val_data, self.val_data_target = self._initialize_data_pair(
            val_data, val_data_target
        )

        # The pred_data_target can be needed to count metrics on the prediction
        self.pred_data, self.pred_data_target = self._initialize_data_pair(
            pred_data, pred_data_target
        )

    def _validate_input_target_type_consistency(
        self,
        input_data: InputType,
        target_data: InputType | None,
    ) -> None:
        """Validate if the input and target data types are consistent.

        Parameters
        ----------
        input_data : InputType
            Input data, can be a path to a folder, a list of paths, or a numpy array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.
        """
        if input_data is not None and target_data is not None:
            if not isinstance(input_data, type(target_data)):
                raise ValueError(
                    f"Inputs for input and target must be of the same type or None. "
                    f"Got {type(input_data)} and {type(target_data)}."
                )
        if isinstance(input_data, list) and isinstance(target_data, list):
            if len(input_data) != len(target_data):
                raise ValueError(
                    f"Inputs and targets must have the same length. "
                    f"Got {len(input_data)} and {len(target_data)}."
                )
            if not isinstance(input_data[0], type(target_data[0])):
                raise ValueError(
                    f"Inputs and targets must have the same type. "
                    f"Got {type(input_data[0])} and {type(target_data[0])}."
                )

    def _list_files_in_directory(
        self,
        input_data,
        target_data=None,
    ) -> tuple[list[Path], list[Path] | None]:
        """List files from input and target directories.

        Parameters
        ----------
        input_data : InputType
            Input data, can be a path to a folder, a list of paths, or a numpy array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.

        Returns
        -------
        (list[Path], Optional[list[Path]])
            A tuple containing lists of file paths for input and target data.
            If target_data is None, the second element will be None.
        """
        input_data = Path(input_data)
        input_files = list_files(input_data, self.data_type, self.extension_filter)
        if target_data is None:
            return input_files, None
        else:
            target_data = Path(target_data)
            target_files = list_files(
                target_data, self.data_type, self.extension_filter
            )
            validate_source_target_files(input_files, target_files)
            return input_files, target_files

    def _convert_paths_to_pathlib(
        self,
        input_data,
        target_data=None,
    ) -> tuple[list[Path], list[Path] | None]:
        """Create a list of file paths from the input and target data.

        Parameters
        ----------
        input_data : InputType
            Input data, can be a path to a folder, a list of paths, or a numpy array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.

        Returns
        -------
        (list[Path], Optional[list[Path]])
            A tuple containing lists of file paths for input and target data.
            If target_data is None, the second element will be None.
        """
        input_files = [
            Path(item) if isinstance(item, str) else item for item in input_data
        ]
        if target_data is None:
            return input_files, None
        else:
            target_files = [
                Path(item) if isinstance(item, str) else item for item in target_data
            ]
            validate_source_target_files(input_files, target_files)
            return input_files, target_files

    def _validate_array_input(
        self,
        input_data: InputType,
        target_data: InputType | None,
    ) -> tuple[Any, Any]:
        """Validate if the input data is a numpy array.

        Parameters
        ----------
        input_data : InputType
            Input data, can be a path to a folder, a list of paths, or a numpy array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.

        Returns
        -------
        (Any, Any)
            A tuple containing the input and target.
        """
        if isinstance(input_data, np.ndarray):
            input_array = [input_data]
            target_array = [target_data] if target_data is not None else None
            return input_array, target_array
        elif isinstance(input_data, list):
            return input_data, target_data
        else:
            raise ValueError(
                f"Unsupported input type for {self.data_type}: {type(input_data)}"
            )

    def _validate_path_input(
        self, input_data: InputType, target_data: InputType | None
    ) -> tuple[list[Path], list[Path] | None]:
        """Validate if the input data is a path or a list of paths.

        Parameters
        ----------
        input_data : InputType
            Input data, can be a path to a folder, a list of paths, or a numpy array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.

        Returns
        -------
        (list[Path], Optional[list[Path]])
            A tuple containing lists of file paths for input and target data.
            If target_data is None, the second element will be None.
        """
        if isinstance(input_data, str | Path):
            if target_data is not None:
                assert isinstance(target_data, str | Path)
            input_list, target_list = self._list_files_in_directory(
                input_data, target_data
            )
            return input_list, target_list
        elif isinstance(input_data, list):
            if target_data is not None:
                assert isinstance(target_data, list)
            input_list, target_list = self._convert_paths_to_pathlib(
                input_data, target_data
            )
            return input_list, target_list
        else:
            raise ValueError(
                f"Unsupported input type for {self.data_type}: {type(input_data)}"
            )

    def _validate_custom_input(self, input_data, target_data) -> tuple[Any, Any]:
        """Convert custom input data to a list of file paths.

        Parameters
        ----------
        input_data : InputType
            Input data, can be a path to a folder, a list of paths, or a numpy array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.

        Returns
        -------
        (Any, Any)
            A tuple containing lists of file paths for input and target data.
            If target_data is None, the second element will be None.
        """
        if self.image_stack_loader is not None:
            return input_data, target_data
        elif isinstance(input_data, str | Path):
            if target_data is not None:
                assert isinstance(target_data, str | Path)
            input_list, target_list = self._list_files_in_directory(
                input_data, target_data
            )
            return input_list, target_list
        elif isinstance(input_data, list):
            if isinstance(input_data[0], str | Path):
                if target_data is not None:
                    assert isinstance(target_data, list)
                input_list, target_list = self._convert_paths_to_pathlib(
                    input_data, target_data
                )
                return input_list, target_list
        else:
            raise ValueError(
                f"If using {self.data_type}, pass a custom "
                f"image_stack_loader or read_source_func"
            )
        return input_data, target_data

    def _initialize_data_pair(
        self,
        input_data: InputType | None,
        target_data: InputType | None,
    ) -> tuple[Any, Any]:
        """
        Initialize a pair of input and target data.

        Parameters
        ----------
        input_data : InputType
            Input data, can be None, a path to a folder, a list of paths, or a numpy
            array.
        target_data : Optional[InputType]
            Target data, can be None, a path to a folder, a list of paths, or a numpy
            array.

        Returns
        -------
        (list of numpy.ndarray or list of pathlib.Path, None or list of numpy.ndarray or
        list of pathlib.Path)
            A tuple containing the initialized input and target data. For file paths,
            returns lists of Path objects. For numpy arrays, returns the arrays
            directly.
        """
        if input_data is None:
            return None, None

        self._validate_input_target_type_consistency(input_data, target_data)

        if self.data_type == SupportedData.ARRAY:
            if isinstance(input_data, np.ndarray):
                return self._validate_array_input(input_data, target_data)
            elif isinstance(input_data, list):
                if isinstance(input_data[0], np.ndarray):
                    return self._validate_array_input(input_data, target_data)
                else:
                    raise ValueError(
                        f"Unsupported input type for {self.data_type}: "
                        f"{type(input_data[0])}"
                    )
            else:
                raise ValueError(
                    f"Unsupported input type for {self.data_type}: {type(input_data)}"
                )
        elif self.data_type in (SupportedData.TIFF, SupportedData.CZI):
            if isinstance(input_data, str | Path):
                return self._validate_path_input(input_data, target_data)
            elif isinstance(input_data, list):
                if isinstance(input_data[0], str | Path):
                    return self._validate_path_input(input_data, target_data)
                else:
                    raise ValueError(
                        f"Unsupported input type for {self.data_type}: "
                        f"{type(input_data[0])}"
                    )
            else:
                raise ValueError(
                    f"Unsupported input type for {self.data_type}: {type(input_data)}"
                )
        elif self.data_type == SupportedData.CUSTOM:
            return self._validate_custom_input(input_data, target_data)
        else:
            raise NotImplementedError(f"Unsupported data type: {self.data_type}")

    def setup(self, stage: str) -> None:
        """
        Setup datasets.

        Lightning hook that is called at the beginning of fit (train + validate),
        validate, test, or predict. Creates the datasets for a given stage.

        Parameters
        ----------
        stage : str
            The stage to set up datasets for.
            Is either 'fit', 'validate', 'test', or 'predict'.

        Raises
        ------
        NotImplementedError
            If stage is not one of "fit", "validate" or "predict".
        """
        if stage == "fit":
            self.train_dataset = create_dataset(
                mode=Mode.TRAINING,
                inputs=self.train_data,
                targets=self.train_data_target,
                config=self.config,
                in_memory=self.use_in_memory,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
            # TODO: ugly, need to find a better solution
            self.stats = self.train_dataset.input_stats
            self.config.set_means_and_stds(
                self.train_dataset.input_stats.means,
                self.train_dataset.input_stats.stds,
                self.train_dataset.target_stats.means,
                self.train_dataset.target_stats.stds,
            )
            self.val_dataset = create_dataset(
                mode=Mode.VALIDATING,
                inputs=self.val_data,
                targets=self.val_data_target,
                config=self.config,
                in_memory=self.use_in_memory,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
        elif stage == "validate":
            self.val_dataset = create_dataset(
                mode=Mode.VALIDATING,
                inputs=self.val_data,
                targets=self.val_data_target,
                config=self.config,
                in_memory=self.use_in_memory,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
            self.stats = self.val_dataset.input_stats
        elif stage == "predict":
            self.predict_dataset = create_dataset(
                mode=Mode.PREDICTING,
                inputs=self.pred_data,
                targets=self.pred_data_target,
                config=self.config,
                in_memory=self.use_in_memory,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
            self.stats = self.predict_dataset.input_stats
        else:
            raise NotImplementedError(f"Stage {stage} not implemented")

    def train_dataloader(self) -> DataLoader:
        """
        Create a dataloader for training.

        Returns
        -------
        DataLoader
            Training dataloader.
        """
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            collate_fn=default_collate,
            **self.config.train_dataloader_params,
        )

    def val_dataloader(self) -> DataLoader:
        """
        Create a dataloader for validation.

        Returns
        -------
        DataLoader
            Validation dataloader.
        """
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            collate_fn=default_collate,
            **self.config.val_dataloader_params,
        )

    def predict_dataloader(self) -> DataLoader:
        """
        Create a dataloader for prediction.

        Returns
        -------
        DataLoader
            Prediction dataloader.
        """
        return DataLoader(
            self.predict_dataset,
            batch_size=self.batch_size,
            collate_fn=default_collate,
            **self.config.test_dataloader_params,
        )



import pytest

from careamics.config.data.ng_data_model import NGDataConfig
from careamics.config.support.supported_patching_strategies import (
    SupportedPatchingStrategy,
)


@pytest.mark.parametrize(
    "patching_strategy",
    [
        {"name": SupportedPatchingStrategy.RANDOM, "patch_size": [16, 16]},
        # {
        #     "name": SupportedPatchingStrategy.SEQUENTIAL,
        #     "patch_size": [16, 16],
        #     "overlap": [4, 4],
        # },
        {
            "name": SupportedPatchingStrategy.TILED,
            "patch_size": [16, 16],
            "overlaps": [4, 4],
        },
        {
            "name": SupportedPatchingStrategy.WHOLE,
        },
    ],
)
def test_ng_data_model_strategy(patching_strategy):

    # Test the DataModel class
    data_model = NGDataConfig(
        data_type="array",
        axes="YX",
        patching=patching_strategy,
    )

    assert data_model.patching.name == patching_strategy["name"]


@pytest.mark.parametrize(
    "axes, patching_strategy",
    [
        ("ZYX", {"name": SupportedPatchingStrategy.RANDOM, "patch_size": [16, 16]}),
        ("YX", {"name": SupportedPatchingStrategy.RANDOM, "patch_size": [16, 16, 16]}),
        (
            "ZYX",
            {
                "name": SupportedPatchingStrategy.TILED,
                "patch_size": [16, 16],
                "overlaps": [4, 4],
            },
        ),
        (
            "SYX",
            {
                "name": SupportedPatchingStrategy.TILED,
                "patch_size": [16, 16, 16],
                "overlaps": [4, 4, 4],
            },
        ),
    ],
)
def test_ng_dataset_invalid_axes_patch(axes, patching_strategy):

    with pytest.raises(ValueError):
        NGDataConfig(
            data_type="array",
            axes=axes,
            patching=patching_strategy,
        )



from collections.abc import Sequence
from enum import Enum
from pathlib import Path
from typing import Any

from numpy.typing import NDArray
from typing_extensions import ParamSpec

from careamics.config.data.ng_data_model import NGDataConfig
from careamics.config.support import SupportedData
from careamics.dataset_ng.patch_extractor import ImageStackLoader, PatchExtractor
from careamics.dataset_ng.patch_extractor.image_stack import (
    CziImageStack,
    GenericImageStack,
    ImageStack,
    InMemoryImageStack,
    ZarrImageStack,
)
from careamics.dataset_ng.patch_extractor.patch_extractor_factory import (
    create_array_extractor,
    create_custom_file_extractor,
    create_custom_image_stack_extractor,
    create_czi_extractor,
    create_ome_zarr_extractor,
    create_tiff_extractor,
)
from careamics.file_io.read import ReadFunc

from .dataset import CareamicsDataset, Mode

P = ParamSpec("P")


# Enum class used to determine which loading functions should be used
class DatasetType(Enum):
    """Labels for the dataset based on the underlying data and how it is loaded."""

    ARRAY = "array"
    IN_MEM_TIFF = "in_mem_tiff"
    LAZY_TIFF = "lazy_tiff"
    IN_MEM_CUSTOM_FILE = "in_mem_custom_file"
    OME_ZARR = "ome_zarr"
    CZI = "czi"
    CUSTOM_IMAGE_STACK = "custom_image_stack"


# bit of a mess of if-else statements
def determine_dataset_type(
    data_type: SupportedData,
    in_memory: bool,
    read_func: ReadFunc | None = None,
    image_stack_loader: ImageStackLoader | None = None,
) -> DatasetType:
    """Determine what the dataset type should be based on the input arguments.

    Parameters
    ----------
    data_type : SupportedData
        The underlying datatype.
    in_memory : bool
        Whether all the data should be loaded into memory. This is argument is ignored
        unless the `data_type` is "tiff" or "custom".
    read_func : ReadFunc, optional
        A function that can be used to load custom data. This argument is
        ignored unless the `data_type` is "custom".
    image_stack_loader : ImageStackLoader, optional
        A function for custom image stack loading. This argument is ignored unless the
        `data_type` is "custom".

    Returns
    -------
    DatasetType
        The Dataset type.

    Raises
    ------
    NotImplementedError
        For lazy-loading (`in_memory=False`) of a custom file type.
    ValueError
        If the `data_type` is "custom" but both `read_func` and `image_stack_loader` are
        None.
    ValueError
        If the `data_type` is unrecognized.
    """
    if data_type == SupportedData.ARRAY:
        # TODO: ignoring in_memory arg, error if False?
        return DatasetType.ARRAY
    elif data_type == SupportedData.TIFF:
        if in_memory:
            return DatasetType.IN_MEM_TIFF
        else:
            return DatasetType.LAZY_TIFF
    elif data_type == SupportedData.CZI:
        return DatasetType.CZI
    elif data_type == SupportedData.CUSTOM:
        if read_func is not None:
            if in_memory:
                return DatasetType.IN_MEM_CUSTOM_FILE
            else:
                raise NotImplementedError(
                    "Lazy loading has not been implemented for custom file types yet."
                )
        elif image_stack_loader is not None:
            # TODO: ignoring im_memory arg
            return DatasetType.CUSTOM_IMAGE_STACK
        else:
            raise ValueError(
                "Found `data_type='custom'` but no `read_func` or `image_stack_loader` "
                "has been provided."
            )
    # TODO: ZARR
    else:
        raise ValueError(f"Unrecognized `data_type`, '{data_type}'.")


# convenience function but should use `create_dataloader` function instead
# For lazy loading custom batch sampler also needs to be set.
def create_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Any,
    targets: Any,
    in_memory: bool,
    read_func: ReadFunc | None = None,
    read_kwargs: dict[str, Any] | None = None,
    image_stack_loader: ImageStackLoader | None = None,
    image_stack_loader_kwargs: dict[str, Any] | None = None,
) -> CareamicsDataset[ImageStack]:
    """
    Convenience function to create the CAREamicsDataset.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.
    in_memory : bool
        Whether all the data should be loaded into memory. This is argument is ignored
        unless the `data_type` in `config` is "tiff" or "custom".
    read_func : ReadFunc, optional
        A function that can that can be used to load custom data. This argument is
        ignored unless the `data_type` in the `config` is "custom".
    read_kwargs : dict of {str, Any}, optional
        Additional key-word arguments to pass to the `read_func`.
    image_stack_loader : ImageStackLoader, optional
        A function for custom image stack loading. This argument is ignored unless the
        `data_type` in the `config` is "custom".
    image_stack_loader_kwargs : {str, Any}, optional
        Additional key-word arguments to pass to the `image_stack_loader`.

    Returns
    -------
    CareamicsDataset[ImageStack]
        The CAREamicsDataset.

    Raises
    ------
    ValueError
        For an unrecognized `data_type` in the `config`.
    """
    data_type = SupportedData(config.data_type)
    dataset_type = determine_dataset_type(
        data_type, in_memory, read_func, image_stack_loader
    )
    if dataset_type == DatasetType.ARRAY:
        return create_array_dataset(config, mode, inputs, targets)
    elif dataset_type == DatasetType.IN_MEM_TIFF:
        return create_tiff_dataset(config, mode, inputs, targets)
    # TODO: Lazy tiff
    elif dataset_type == DatasetType.CZI:
        return create_czi_dataset(config, mode, inputs, targets)
    elif dataset_type == DatasetType.IN_MEM_CUSTOM_FILE:
        if read_kwargs is None:
            read_kwargs = {}
        assert read_func is not None  # should be true from `determine_dataset_type`
        return create_custom_file_dataset(
            config, mode, inputs, targets, read_func=read_func, read_kwargs=read_kwargs
        )
    elif dataset_type == DatasetType.CUSTOM_IMAGE_STACK:
        if image_stack_loader_kwargs is None:
            image_stack_loader_kwargs = {}
        assert image_stack_loader is not None  # should be true
        return create_custom_image_stack_dataset(
            config,
            mode,
            inputs,
            targets,
            image_stack_loader,
            **image_stack_loader_kwargs,
        )
    else:
        raise ValueError(f"Unrecognized dataset type, {dataset_type}.")


def create_array_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Sequence[NDArray[Any]],
    targets: Sequence[NDArray[Any]] | None,
) -> CareamicsDataset[InMemoryImageStack]:
    """
    Create a CAREamicsDataset from array data.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.

    Returns
    -------
    CareamicsDataset[InMemoryImageStack]
        A CAREamicsDataset.
    """
    input_extractor = create_array_extractor(source=inputs, axes=config.axes)
    target_extractor: PatchExtractor[InMemoryImageStack] | None
    if targets is not None:
        target_extractor = create_array_extractor(source=targets, axes=config.axes)
    else:
        target_extractor = None
    return CareamicsDataset(config, mode, input_extractor, target_extractor)


def create_tiff_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Sequence[Path],
    targets: Sequence[Path] | None,
) -> CareamicsDataset[InMemoryImageStack]:
    """
    Create a CAREamicsDataset from tiff files that will be all loaded into memory.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.

    Returns
    -------
    CareamicsDataset[InMemoryImageStack]
        A CAREamicsDataset.
    """
    input_extractor = create_tiff_extractor(
        source=inputs,
        axes=config.axes,
    )
    target_extractor: PatchExtractor[InMemoryImageStack] | None
    if targets is not None:
        target_extractor = create_tiff_extractor(source=targets, axes=config.axes)
    else:
        target_extractor = None
    dataset = CareamicsDataset(config, mode, input_extractor, target_extractor)
    return dataset


def create_czi_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Sequence[Path],
    targets: Sequence[Path] | None,
) -> CareamicsDataset[CziImageStack]:
    """
    Create a dataset from CZI files.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.

    Returns
    -------
    CareamicsDataset[CziImageStack]
        A CAREamicsDataset.
    """

    input_extractor = create_czi_extractor(source=inputs, axes=config.axes)
    target_extractor: PatchExtractor[CziImageStack] | None
    if targets is not None:
        target_extractor = create_czi_extractor(source=targets, axes=config.axes)
    else:
        target_extractor = None
    dataset = CareamicsDataset(config, mode, input_extractor, target_extractor)
    return dataset


def create_ome_zarr_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Sequence[Path],
    targets: Sequence[Path] | None,
) -> CareamicsDataset[ZarrImageStack]:
    """
    Create a dataset from OME ZARR files.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.

    Returns
    -------
    CareamicsDataset[ZarrImageStack]
        A CAREamicsDataset.
    """

    input_extractor = create_ome_zarr_extractor(source=inputs, axes=config.axes)
    target_extractor: PatchExtractor[ZarrImageStack] | None
    if targets is not None:
        target_extractor = create_ome_zarr_extractor(source=targets, axes=config.axes)
    else:
        target_extractor = None
    dataset = CareamicsDataset(config, mode, input_extractor, target_extractor)
    return dataset


def create_custom_file_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Sequence[Path],
    targets: Sequence[Path] | None,
    *,
    read_func: ReadFunc,
    read_kwargs: dict[str, Any],
) -> CareamicsDataset[InMemoryImageStack]:
    """
    Create a CAREamicsDataset from custom files that will be all loaded into memory.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.
    read_func : Optional[ReadFunc], optional
        A function that can that can be used to load custom data. This argument is
        ignored unless the `data_type` is "custom".
    image_stack_loader : Optional[ImageStackLoader], optional
        A function for custom image stack loading. This argument is ignored unless the
        `data_type` is "custom".

    Returns
    -------
    CareamicsDataset[InMemoryImageStack]
        A CAREamicsDataset.
    """
    input_extractor = create_custom_file_extractor(
        source=inputs, axes=config.axes, read_func=read_func, read_kwargs=read_kwargs
    )
    target_extractor: PatchExtractor[InMemoryImageStack] | None
    if targets is not None:
        target_extractor = create_custom_file_extractor(
            source=targets,
            axes=config.axes,
            read_func=read_func,
            read_kwargs=read_kwargs,
        )
    else:
        target_extractor = None
    dataset = CareamicsDataset(config, mode, input_extractor, target_extractor)
    return dataset


def create_custom_image_stack_dataset(
    config: NGDataConfig,
    mode: Mode,
    inputs: Any,
    targets: Any | None,
    image_stack_loader: ImageStackLoader[P, GenericImageStack],
    *args: P.args,
    **kwargs: P.kwargs,
) -> CareamicsDataset[GenericImageStack]:
    """
    Create a CAREamicsDataset from a custom `ImageStack` class.

    The custom `ImageStack` class can be loaded using the `image_stack_loader` function.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    mode : Mode
        Whether to create the dataset in "training", "validation" or "predicting" mode.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.
    image_stack_loader : ImageStackLoader
        A function for custom image stack loading. This argument is ignored unless the
        `data_type` is "custom".
    *args : Any
        Positional arguments to pass to the `image_stack_loader`.
    **kwargs : Any
        Key-word arguments to pass to the `image_stack_loader`.

    Returns
    -------
    CareamicsDataset[GenericImageStack]
        A CAREamicsDataset.
    """
    input_extractor = create_custom_image_stack_extractor(
        inputs,
        config.axes,
        image_stack_loader,
        *args,
        **kwargs,
    )
    target_extractor: PatchExtractor[GenericImageStack] | None
    if targets is not None:
        target_extractor = create_custom_image_stack_extractor(
            targets,
            config.axes,
            image_stack_loader,
            *args,
            **kwargs,
        )
    else:
        target_extractor = None
    dataset = CareamicsDataset(config, mode, input_extractor, target_extractor)
    return dataset



"""Patching strategies Pydantic models."""

__all__ = [
    "RandomPatchingModel",
    "SequentialPatchingModel",
    "TiledPatchingModel",
    "WholePatchingModel",
]


from .random_patching_model import RandomPatchingModel
from .sequential_patching_model import SequentialPatchingModel
from .tiled_patching_model import TiledPatchingModel
from .whole_patching_model import WholePatchingModel
